#ifndef ONI_H
#error Do not include this file directly, include Oni/Oni.h instead!
#endif

#ifndef SYMBOLS_FUNC_H
#define SYMBOLS_FUNC_H


#define DefFunc( type, name, callingconvention, args, address) typedef type ( callingconvention * _##name ) args; static const _##name name = (_##name)address

// Orig timing functions, replaced by DDrMachineTime_* if option "usegettickcount" is enabled
DefFunc(int64_t, UUrMachineTime_High, ONICALL, (), 0x00426480);
DefFunc(double, UUrMachineTime_High_Frequency, ONICALL, (), 0x004264b0);
DefFunc(int64_t, UUrMachineTime_Sixtieths, ONICALL, (), 0x004263e0);

// Code to initialize platform stuff (like disabling system key combination Alt+Tab)
DefFunc(void, UUrPlatform_Initialize, ONICALL, (), 0x00426010);
DefFunc(void, UUrPlatform_Terminate, ONICALL, (), 0x00426310);

// Oni's file access methods
DefFunc(FILE*, oni_fopen, __cdecl, (const char*, const char*), 0x0051ea9f);
DefFunc(int, oni_fprintf, __cdecl, (FILE*, const char*, ...), 0x0051ebbf);
DefFunc(int, oni_fflush, __cdecl, (FILE*), 0x0051eab2);

// Oni's main method?
DefFunc(void, ONiMain, __cdecl, (int ArgCount, char *ArgList[]), 0x004d3280);

DefFunc(uint16_t, CLrGetCommandLine, ONICALL, (int inArgc, const char* inArgv[], char*** outArgV), 0x0043ffd0);

// Initialization code, used to hook in windowed modes
DefFunc(short, ONrPlatform_Initialize, ONICALL, (ONtPlatformData *PlatformData), 0x0050f670);
DefFunc(LRESULT, ONrPlatform_WindowProc, CALLBACK, (HWND Window, UINT Message, WPARAM WParam, LPARAM LParam), 0x0050f7a0);

// Retrieves a list of resolutions, replaced by "daodan_enumerate_valid_display_modes"
DefFunc(unsigned int, gl_enumerate_valid_display_modes, ONICALL, (M3tDisplayMode modes[16]), 0x004083a0);
DefFunc(int, gl_platform_set_pixel_format, ONICALL, (HDC hdc), 0x00407b50);

// Original graphics initialization/cleanup functions patched to use
// DD_GLrPlatform_Initialize/DD_GLrPlatform_Dispoose instead if daodangl is used.
DefFunc(UUtBool, gl_platform_initialize, ONICALL, (void), 0x00407da0);
DefFunc(void, gl_platform_dispose, ONICALL, (void), 0x00408210);

// Retrieve gamma value from settings
DefFunc(float, ONrPersist_GetGamma, ONICALL, (), 0x0050f450);

// Retrieve if the game was completed
DefFunc(uint8_t, ONrPersist_GetWonGame, ONICALL, (), 0x0050f660);

// Print a message to the startup.txt log
DefFunc(void, UUrStartupMessage, __cdecl, (const char* fmt, ...), 0x00424860);

// Probably to spawn a new character
DefFunc(int16_t, ONrGameState_NewCharacter, ONICALL, (void* CHAR, void* AISA, void* flag, uint32_t* list_location), 0x004dac50);

// Remove character
DefFunc(void, ONrGameState_DeleteCharacter, ONICALL, (Character *inCharacter), 0x004DC480);

DefFunc(void, ONrGameState_Timer_Start, ONICALL, (char* function, int time), 0x004FD370);
DefFunc(void, ONrGameState_HandleUtilityInput, ONICALL, (const void* inInput), 0x004f6200);
DefFunc(uint8_t, ONrDebugKey_WentDown, ONICALL, (uint32_t inKey), 0x005050d0);

// Get active (player?) character
DefFunc(ActiveCharacter*, ONrGetActiveCharacter, ONICALL, (void* CharacterPtr), 0x004f1180);

DefFunc(uint32_t, ONrCharacter_GetHealthShade, ONICALL, (uint32_t health, uint32_t maxhealth), 0x004EF450);
DefFunc(void, ONrCharacter_NewAnimationHook, ONICALL, (Character *ioCharacter, ActiveCharacter *ioActiveCharacter), 0x004E97A0);
DefFunc(void, ONrCharacter_SetAnimationExternal, ONICALL, (Character *ioCharacter, short state, void* animation, int interpolation), 0x004EB340);
DefFunc(void, ONrCharacter_SetHitPoints, ONICALL, (Character *ioCharacter, uint32_t inHitPoints), 0x004EB220);
DefFunc(void, ONrCharacter_SetCharacterClass, ONICALL, (Character* Char, ONCC* Class), 0x004D7C30);
DefFunc(void, ONrCorpse_Create, ONICALL, (Character* Character), 0x004EF340);


// Print message to console
DefFunc(void, COrTextArea_Print, ONICALL, (void* area, uint32_t priority, uint32_t textshade, uint32_t textshadowshade, const char* text, uint32_t unk_alwaws_0, uint32_t fadetime), 0x00431340);
DefFunc(int16_t, COrTextArea_Resize, ONICALL, (void* inTextArea, UUtRect* inBounds, int16_t inNumTextEntries), 0x00431460);

// Original cheat function
DefFunc(uint8_t, ONrCheater, ONICALL, (uint32_t cheat), 0x004f5c30);

DefFunc(int16_t, ONrMechanics_Register, ONICALL, (uint32_t inObjectType, uint32_t inObjectTypeIndex, char* inGroupName,
	uint32_t inSizeInMemory, OBJtMethods* inObjectMethods, uint32_t inFlags, void* inMechanicsMethods), 0x00507450);
DefFunc(int16_t, OBJrTriggerVolume_IntersectsCharacter, ONICALL, (const OBJtObject* inObject, uint32_t inMask, const void* inCharacter), 0x004cc680);


// ?
DefFunc(int, AUrMessageBox, __cdecl, (int Buttons, char *Message, ...), 0x004378c0);
DefFunc(char*, SSrMessage_Find, ONICALL, (char* message_key), 0x0047F550);
DefFunc(void, ONiGameState_FindAutoPromptMessage, ONICALL, (char* Note, void* ptr), 0x004FDBE0);

// Register a BSL function with a return type
DefFunc(uint16_t, SLrScript_Command_Register_ReturnType, ONICALL, (char* name, char* desc, char* argfmt, sl_type type, sl_func callback), 0x00477b20);

// Register a BSL function without a return value
DefFunc(uint16_t, SLrScript_Command_Register_Void, ONICALL, (char* name, char* desc, char* argfmt, sl_func callback), 0x00477b40);

// Register a global BSL variable
DefFunc(uint16_t, SLrGlobalVariable_Register_Bool, ONICALL, (char* name, char* desc, UUtBool* data), 0x00477f50);
DefFunc(uint16_t, SLrGlobalVariable_Register_Int32, ONICALL, (char* name, char* desc, int32_t* data), 0x00477e30);
DefFunc(uint16_t, SLrGlobalVariable_Register_Float, ONICALL, (char* name, char* desc, float* data), 0x00477ec0);
DefFunc(uint16_t, SLrGlobalVariable_Register_String, ONICALL, (char* name, char* desc, char* data), 0x00477fe0);

// Print message to console ?
DefFunc(int, COrMessage_Print, ONICALL, (const char* Message, const char* Key, uint32_t fadeTime), 0x004304B0);
DefFunc(UUtBool, COrCommand_Execute, ONICALL, (char* command), 0x004317D0);

DefFunc(void, COrConsole_StatusLine_Display, ONICALL, (), 0x00431E70);

// Get pointer to specified data instance
DefFunc(int16_t, TMrInstance_GetDataPtr, ONICALL, (int tag, char* name, void* pointer), 0x004232E0);

// Get name of data instance pointed to
DefFunc(char*, TMrInstance_GetInstanceName, ONICALL, (void* InstancePointer), 0x00423D90);

DefFunc(short, TMrInstance_GetDataPtr_ByNumber, ONICALL, (int tag, int number, void** out), 0x00423680);
DefFunc(uint32_t, TMrInstance_GetTagCount, ONICALL, (int tag), 0x004236F0);

// Draw text on the screen
DefFunc(int16_t, TSrContext_DrawText, ONICALL, (void* TSrContext, char* Text, char alpha, uint32_t usuallyzero, void* pRect), 0x0042DF00);
DefFunc(int16_t, TSrContext_New, ONICALL, (void* FontInstance, int size, int hthsik1,int hthsik2,int hthsik3, void* TSrContext), 0x0042EA30);
DefFunc(int16_t, TSrContext_SetFontSize, ONICALL, (void* ioTextContext, uint16_t inSize), 0x0042ED50);
DefFunc(int16_t, TSrContext_SetShade, ONICALL, (void* ioTextContext, uint32_t inShade), 0x0042EE50);

DefFunc(uint16_t, TRrAnimation_GetDuration, ONICALL, (void* Animation), 0x00428740);
DefFunc(uint16_t, TRrAnimation_GetTo, ONICALL, (void* Animation), 0x00428730);
DefFunc(uint16_t, TRrAnimation_GetFrom, ONICALL, (void* Animation), 0x00428720);

DefFunc(uint16_t, iSetCharacterClass, ONICALL, (sl_callinfo* callinfo, uint32_t numargs, sl_arg args[], int* dontuse1, int* dontuse2, sl_arg* ret), 0x004D99D0);
DefFunc(uint16_t, AI2iScript_Spawn, ONICALL, (sl_callinfo* callinfo, uint32_t numargs, sl_arg args[], int* dontuse1, int* dontuse2, sl_arg* ret), 0x004B4780);
DefFunc(void, AI2rSmite, ONICALL, (Character *inCharacter, UUtBool inSpareCharacter), 0x0048BDA0);

DefFunc(int, OBJrObjectType_EnumerateObjects, ONICALL, (int inObjectType, OBJtEnumCallback_Object inEnumCallback, int inUserData), 0x004D0080);

DefFunc(void, OBJrDoor_Open, ONICALL, (DoorObject *inObject, Character *inCharacter), 0x004C26C0);
DefFunc(void, OBJrDoor_ForceOpen, ONICALL, (short id), 0x004C1EE0);
DefFunc(short, OBJrConsole_OnActivate, ONICALL, (void *inObject, Character *inCharacter), 0x004C0880);
DefFunc(void*, OBJrConsole_GetByID, ONICALL, (short ID), 0x004C0950);


DefFunc(void, ONiDrawWeaponSight, ONICALL, (Character* Char), 0x004E1900);
DefFunc(void, AI2rDisplayDebuggingInfo, ONICALL, (Character* Char), 0x0048C5F0);
DefFunc(uint32_t, M3rTextureMap_New, ONICALL, (short width, short height, int type, int allocated, int flags, char* name, void** output), 0x0041EB00);
DefFunc(void, M3rDraw_BigBitmap, ONICALL, (M3tTextureMap_Big* inBigBitmap, const M3tPointScreen *inDestPoint, UUtUns16 inWidth, UUtUns16 inHeight, UUtUns32 inShade, UUtUns16 inAlpha), 0x0041f6e0);
DefFunc(void, M3rDraw_Bitmap, ONICALL, (void* inBitmap, const M3tPointScreen* inDestPoint, UUtUns16 inWidth, UUtUns16 inHeight, UUtUns32 inShade, UUtUns16 inAlpha), 0x0041f4a0);

DefFunc(void, M3rGeom_Line_Light, ONICALL, (const M3tPoint3D *point1, const M3tPoint3D *point2, uint32_t shade), 0x0041f440);

// Make a dialog element (in)visible and change its position
DefFunc(void, WMrWindow_SetVisible, ONICALL, (void* window, int visibility), 0x00475a10);
DefFunc(void, WMrWindow_SetLocation, ONICALL, (void* window, int x, int y), 0x004756d0);

// Plays movie using Bink blitting method (ie., non-OpenGL blitting).
DefFunc(void, ONrMovie_Play, ONICALL, (const char* movie, int unknown), 0x004d5580);

// Sets minimum and maximum values for a slider control.
DefFunc(void, WMrSlider_SetRange, ONICALL, (void* window, int min_value, int max_value), 0x00472650);

// Enables or disables window.
DefFunc(void, WMrWindow_SetEnabled, ONICALL, (void* window, UUtBool enabled), 0x00475580);

// Sets gamma to specified factor (0.0 .. 1.0).
DefFunc(void, M3rSetGamma, ONICALL, (float factor), 0x00407a60);

// Resets gl_api pointers, unloads OPENGL32.DLL.
DefFunc(void, gl_unload_library, ONICALL, (void), 0x0040ac40);

// Notifies platform-specific local input implementation about internal (ie.,
// real) input mode change.
DefFunc(void, LIrPlatform_Mode_Set, ONICALL, (unsigned int active_mode), 0x00402cf0);

// Terminates platform-specific local-input.
DefFunc(void, LIrPlatform_Terminate, ONICALL, (void), 0x00403620);

// Key_config callback for each mapping found
DefFunc(uint16_t, LIrBinding_Add, ONICALL, (uint32_t key, const char* name), 0x00403c60);

// Called when processing key input
DefFunc(void, LIrActionBuffer_Add, ONICALL, (void* unknown, LItDeviceInput* input), 0x00403b30);

// Called during the game loop to run the Windows message loop
DefFunc(UUtBool, LIiPlatform_InputEvent_GetEvent, ONICALL, (void), 0x004036d0);

// Called by the game loop to get the latest input frames
DefFunc(void, LIrActionBuffer_Get, ONICALL, (short* count, LItActionBuffer **buffers), 0x00403be0);

// Gets the mouse position
DefFunc(void, LIrPlatform_InputEvent_GetMouse, ONICALL, (int active, LItInputEvent *info), 0x00402ca0);

// Checks if a keyboard key is pressed
DefFunc(UUtBool, LIrPlatform_TestKey, ONICALL, (int key, int active), 0x00403930);

// Translates a key name (from key_config.txt) to a key code
DefFunc(int, LIrTranslate_InputName, ONICALL, (char *name), 0x00403a90);

// Updates TargetGameTime in the game state
DefFunc(void, ONrGameState_UpdateServerTime, ONICALL, (GameState *game_state), 0x004fbeb0);

#undef DefFunc

#endif
